/***************************************************************************
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 dated June, 1991.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 */

#include <stdio.h>
#include <string.h>
#include <netdb.h>
#include <upnp/upnp.h>
#include <upnp/ixml.h>
#include <sys/types.h>
#include <sys/socket.h>

/* #define DEBUG 1 */

char *ip_address;
unsigned short port;
IXML_NodeList *radios;

int read_config(void);
int radio_get_info (const char *filename, struct File_Info *info);
UpnpWebFileHandle radio_open (const char *filename, enum UpnpOpenFileMode Mode);
int radio_read (UpnpWebFileHandle fileHnd, char *buf, size_t buflen);
int radio_write (UpnpWebFileHandle fileHnd, char *buf, size_t buflen);
int radio_seek (UpnpWebFileHandle fileHnd, long offset, int origin);
int radio_close (UpnpWebFileHandle fileHnd);

struct UpnpVirtualDirCallbacks callbacks =
{ radio_get_info, radio_open, radio_read, radio_write, radio_seek, radio_close };

char *
GetElementValue( IN IXML_Element * element )
{
	IXML_Node *child = ixmlNode_getFirstChild( ( IXML_Node * ) element );
	char *temp = NULL;

	if( ( child != 0 ) && (ixmlNode_getNodeType( child ) == eTEXT_NODE ))
		temp = strdup( ixmlNode_getNodeValue( child ) );

	return temp;
}

int read_config(void)
{
	IXML_Document *config;
	IXML_NodeList *nodelist;
	IXML_Element *ipaddr;
	IXML_Element *portnr;

	config = ixmlLoadDocument("/etc/mserv/radio.xml");
	if (!config) {
#ifdef DEBUG
		printf("Parse of radio.xml failed\n");
#endif
		return -1;
	}
	radios = ixmlDocument_getElementsByTagName(config, "radio");
	if (!radios) return -1;

	ip_address = NULL;
	port = 8080;

	if (nodelist = ixmlDocument_getElementsByTagName(config, "ipaddr")) {
		ipaddr = (IXML_Element *) ixmlNodeList_item(nodelist, 0);
		ip_address = GetElementValue(ipaddr);
	}
	if (nodelist = ixmlDocument_getElementsByTagName(config, "port")) {
		portnr = (IXML_Element *) ixmlNodeList_item(nodelist, 0);
		port = atoi(GetElementValue(portnr));
	}

	return 0;
}

int sock_read_line(int sock, char *buff, const int len)
{
	char c = '\0';
        int read_bytes, pos;

	if (!buff) {
		return 0;
	} else if (len <= 0) {
		return 0;
	}

	pos = 0;
	read_bytes = recv(sock, &c, 1, 0);

	if (read_bytes < 0) {
		return 0;
	}

	while ((c != '\n') && (pos < len) && (read_bytes == 1)) {
		if (c != '\r')
			buff[pos++] = c;
		read_bytes = recv(sock, &c, 1, 0);
	}

	if (read_bytes == 1) {
		buff[pos] = '\0';
		return 1;
	} else {
		return 0;
	}
}

int radio_connect (const char *url)
{
	char *host;
	int   port, handle;
	char *path, *p;
	struct hostent *hostent;
	struct in_addr addr;
	struct sockaddr_in server;
	char buf[200];

#ifdef DEBUG
	printf("Connect to %s\n", url);
#endif
	handle = -1;
	if (strncmp(url, "http://", 7))
		return -1;

	port = 80;

	host = strdup(url + 7);
	p = index(host, '/');
	if (!p)
		path=strdup("/");
	else {
		path=strdup(p);
		*p = 0;
	}
	p = index(host, ':');
	if (p) {
		port = atoi(p+1);
		*p = 0;
	}
#ifdef DEBUG
	printf("Host: %s, Port: %d, Path: %s\n", host, port, path);
#endif

	memset(&server, 0, sizeof(struct sockaddr_in));
	if (inet_aton(host, &addr)) {
		memcpy(&server.sin_addr.s_addr,&addr.s_addr,sizeof(server.sin_addr.s_addr));
	} else {
		hostent = gethostbyname(host);
		if (hostent) {
			memcpy(&server.sin_addr.s_addr,&(*(struct in_addr *)hostent->h_addr).s_addr,sizeof(server.sin_addr.s_addr));
		}
	}

	handle = socket (AF_INET, SOCK_STREAM, 0);
	if (handle == -1)
		return -1;


	server.sin_family = AF_INET;
	server.sin_port = htons(port);
	
	connect(handle, (struct sockaddr *)&server, sizeof(server));

	snprintf(buf, sizeof(buf), "GET %s HTTP/1.0\r\nHost: %s\r\n\r\n", path, host);
	send(handle, buf, sizeof(buf), 0);


	while(strcmp(buf,"")) {
		sock_read_line(handle, buf, sizeof(buf));
#ifdef DEBUG
		printf("%s\n", buf);
#endif
	}
	return handle;
}

int radio_get_info (const char *filename, struct File_Info *info)
{
#ifdef DEBUG
	printf("info %s\n", filename);
#endif
	info->file_length=-1;
	info->last_modified=0;
	info->is_directory=0;
	info->is_readable=1;
	info->content_type=strdup("audio/mp3");

	return 0;
}

UpnpWebFileHandle radio_open (const char *filename, enum UpnpOpenFileMode Mode)
{
	char *num;
	int n, handle;
	IXML_Element *element;
	IXML_NamedNodeMap *attrs;
#ifdef DEBUG
	printf("open %s\n", filename);
#endif
	if (Mode != UPNP_READ)
		return (UpnpWebFileHandle) -1;

	num = index(filename + 1, '/') + 1;
	n = atoi(num);

#ifdef DEBUG
	printf("open index %d\n", n);
#endif
	if (n >= ixmlNodeList_length(radios))
		return (UpnpWebFileHandle) -1;

	element = (IXML_Element *) ixmlNodeList_item(radios, n);

	handle = radio_connect(ixmlElement_getAttribute(element, "url"));
	return (UpnpWebFileHandle) handle;
}

int radio_read (UpnpWebFileHandle fileHnd, char *buf, size_t buflen)
{
	int handle;

#ifdef DEBUG
	printf("read\n");
#endif
	handle = (int) fileHnd;
	return recv(handle, buf, buflen, 0);
}

int radio_write (UpnpWebFileHandle fileHnd, char *buf, size_t buflen)
{
#ifdef DEBUG
	printf("write\n");
#endif
	return -1;
}

int radio_seek (UpnpWebFileHandle fileHnd, long offset, int origin)
{
#ifdef DEBUG
	printf("seek\n");
#endif
	return -1;
}

int radio_close (UpnpWebFileHandle fileHnd)
{
	int handle;

	handle = (int) fileHnd;
	return close(handle);
}

char * GetFirstDocumentItem(IXML_Document * doc, const char *item )
{
	IXML_NodeList *nodeList = NULL;
	IXML_Node *textNode = NULL;
	IXML_Node *tmpNode = NULL;

	char *ret = NULL;
	nodeList = ixmlDocument_getElementsByTagName( doc, ( char * )item );

	if( nodeList ) {
		if( ( tmpNode = ixmlNodeList_item( nodeList, 0 ) ) ) {
			textNode = ixmlNode_getFirstChild( tmpNode );
			if (ixmlNode_getNodeValue( textNode ) )
				ret = strdup( ixmlNode_getNodeValue( textNode ) );
		}
	}

	if( nodeList )
		ixmlNodeList_free( nodeList );

	return ret;
}

int browse(IXML_Document *in, IXML_Document **out)
{
	int ret;
	IXML_Document *doc;
	IXML_Element  *element, *element1, *element2, *element3;
	IXML_CDATASection *cdata;
	DOMString      p;
	char *oid, *flag, *filter, *sort, *indexp, *countp;
	char buf[200];
	int index, count, n;

	oid    = GetFirstDocumentItem(in, "ObjectID");
	flag   = GetFirstDocumentItem(in, "BrowseFlag"); 
	filter = GetFirstDocumentItem(in, "Filter");
	indexp = GetFirstDocumentItem(in, "StartingIndex");
	countp = GetFirstDocumentItem(in, "RequestedCount");
	sort   = GetFirstDocumentItem(in, "SortCriteria");

	count = atoi(countp);
	index = atoi(indexp);

#ifdef DEBUG
	printf("OID=%s, flag=%s, filter=%s, index=%d, count=%d\n",
		oid, flag, filter, index, count);
#endif
	doc = ixmlDocument_createDocument();

	element = ixmlDocument_createElement(doc, "DIDL-Lite");
	ixmlElement_setAttribute(element, "xmlns", "urn:schemas-upnp-org:metadata-1-0/DIDL-Lite");
	ixmlElement_setAttribute(element, "xmlns:dc", "http://purl.org/dc/elements/1.1/");
	ixmlElement_setAttribute(element, "xmlns:upnp", "urn:schemas-upnp-org:metadata-1-0/upnp/");
	ixmlNode_appendChild((IXML_Node*)doc,(IXML_Node*)element);

	n = index;
	if (!strcmp(oid,"0")) { /* Only root can have childs */
		while ((n < ixmlNodeList_length(radios)) && ((n-index) < count)) {
			element1 = ixmlDocument_createElement(doc, "container");
			snprintf(buf, sizeof(buf), "media%d", n);
			ixmlElement_setAttribute(element1, "id", buf);
			ixmlElement_setAttribute(element1, "parentID", "0");
			ixmlElement_setAttribute(element1, "restricted", "true");
			ixmlElement_setAttribute(element1, "childCount", "0");
			ixmlNode_appendChild((IXML_Node*)element,(IXML_Node*)element1);

			element2 = ixmlDocument_createElement(doc, "dc:title");
			ixmlNode_appendChild((IXML_Node*)element1,(IXML_Node*)element2);

			element3 = (IXML_Element*)ixmlNodeList_item(radios, n);
			cdata = ixmlDocument_createCDATASection(doc, GetElementValue(element3));
			ixmlNode_appendChild((IXML_Node*)element2,(IXML_Node*)cdata);

			element2 = ixmlDocument_createElement(doc, "dc:creator");
			ixmlNode_appendChild((IXML_Node*)element1,(IXML_Node*)element2);

			cdata = ixmlDocument_createCDATASection(doc, "mserv");
			ixmlNode_appendChild((IXML_Node*)element2,(IXML_Node*)cdata);

			element2 = ixmlDocument_createElement(doc, "upnp:class");
			ixmlNode_appendChild((IXML_Node*)element1,(IXML_Node*)element2);

			cdata = ixmlDocument_createCDATASection(doc, "object.item.audioItem.musicTrack");
			ixmlNode_appendChild((IXML_Node*)element2,(IXML_Node*)cdata);
			element2 = ixmlDocument_createElement(doc, "res");
			ixmlElement_setAttribute(element1, "protocolInfo", "http-get:*:audio/mp3:*");
			ixmlNode_appendChild((IXML_Node*)element1,(IXML_Node*)element2);
			snprintf(buf, sizeof(buf), "http://%s:%d/media/%d.mp3", ip_address, port, n);
			cdata = ixmlDocument_createCDATASection(doc, buf);
			ixmlNode_appendChild((IXML_Node*)element2,(IXML_Node*)cdata);
			n++;
		}
	}
	p = ixmlDocumenttoString(doc);
	ret = UpnpAddToActionResponse(out, "Browse", "urn:upnp-org:serviceId:ContentDirectory", "Result", p);
	snprintf(buf, sizeof(buf), "%d", n-index);
	ret = UpnpAddToActionResponse(out, "Browse", "urn:upnp-org:serviceId:ContentDirectory", "NumberReturned", buf);
	snprintf(buf, sizeof(buf), "%d", ixmlNodeList_length(radios));
	ret = UpnpAddToActionResponse(out, "Browse", "urn:upnp-org:serviceId:ContentDirectory", "TotalMatches", buf);
	ret = UpnpAddToActionResponse(out, "Browse", "urn:upnp-org:serviceId:ContentDirectory", "UpdateID", "0");
	if (ret != UPNP_E_SUCCESS) {
		printf("Error in UpnpAddToActionResponse(): %d\n", ret);
	}
	free(oid);
	free(flag);
	free(filter);
	free(sort);
	free(indexp);
	free(countp);
	return ret;
}

int action(struct Upnp_Action_Request *ca_event)
{
	int action_found = 0;

	ca_event->ErrCode = 0;
	ca_event->ActionResult = NULL;

#ifdef DEBUG
	printf("%s/%s:%s\n", ca_event->DevUDN, ca_event->ServiceID, ca_event->ActionName);
#endif
	if (strcmp(ca_event->DevUDN, "uuid:52c2a23a-299e-460e-9b43-7987ff821e4b"))
		return 1;

	if (strcmp(ca_event->ServiceID, "urn:upnp-org:serviceId:ContentDirectory") == 0) {
		if (strcmp(ca_event->ActionName, "Browse") == 0) {
			browse(ca_event->ActionRequest, &ca_event->ActionResult);
			action_found=1;
		}
	}
	if (action_found)
		ca_event->ErrCode = UPNP_E_SUCCESS;

	return ca_event->ErrCode;
}

int event(Upnp_EventType event_type, void *event, void *cookie)
{
	switch (event_type) {
		case UPNP_DISCOVERY_ADVERTISEMENT_ALIVE:
#ifdef DEBUG
			printf("UPNP_DISCOVERY_ADVERTISEMENT_ALIVE\n");
#endif
			break;

		case UPNP_DISCOVERY_ADVERTISEMENT_BYEBYE:
#ifdef DEBUG
			printf("UPNP_DISCOVERY_ADVERTISEMENT_BYEBYE\n");
#endif
			break;

		case UPNP_DISCOVERY_SEARCH_RESULT:
#ifdef DEBUG
			printf("UPNP_DISCOVERY_SEARCH_RESULT\n");
#endif
			break;

		case UPNP_DISCOVERY_SEARCH_TIMEOUT:
#ifdef DEBUG
			printf("UPNP_DISCOVERY_SEARCH_TIMEOUT\n");
#endif
			break;

		case UPNP_CONTROL_ACTION_REQUEST:
#ifdef DEBUG
			printf("UPNP_CONTROL_ACTION_REQUEST\n");
#endif
			action((struct Upnp_Action_Request *)event);
			break;

	        case UPNP_CONTROL_ACTION_COMPLETE:
#ifdef DEBUG
			printf("UPNP_CONTROL_ACTION_COMPLETE\n");
#endif
			break;

		case UPNP_CONTROL_GET_VAR_REQUEST:
#ifdef DEBUG
			printf("UPNP_CONTROL_GET_VAR_REQUEST\n");
#endif
			break;

		case UPNP_CONTROL_GET_VAR_COMPLETE:
#ifdef DEBUG
			printf("UPNP_CONTROL_GET_VAR_COMPLETE\n");
#endif
			break;

		case UPNP_EVENT_SUBSCRIPTION_REQUEST:
#ifdef DEBUG
			printf("UPNP_EVENT_SUBSCRIPTION_REQUEST\n");
#endif
			break;

		case UPNP_EVENT_RECEIVED:
#ifdef DEBUG
			printf("UPNP_EVENT_RECEIVED\n");
#endif
			break;

		case UPNP_EVENT_RENEWAL_COMPLETE:
#ifdef DEBUG
			printf("UPNP_EVENT_RENEWAL_COMPLETE\n");
#endif
			break;

		case UPNP_EVENT_SUBSCRIBE_COMPLETE:
#ifdef DEBUG
			printf("UPNP_EVENT_SUBSCRIBE_COMPLETE\n");
#endif
			break;

		case UPNP_EVENT_UNSUBSCRIBE_COMPLETE:
#ifdef DEBUG
			printf("UPNP_EVENT_UNSUBSCRIBE_COMPLETE\n");
#endif
			break;

		case UPNP_EVENT_AUTORENEWAL_FAILED:
#ifdef DEBUG
			printf("UPNP_EVENT_AUTORENEWAL_FAILED\n");
#endif
			break;

		case UPNP_EVENT_SUBSCRIPTION_EXPIRED:
#ifdef DEBUG
			printf("UPNP_EVENT_SUBSCRIPTION_EXPIRED\n");
#endif
			break;

		default:
#ifdef DEBUG
			printf("other event\n");
#endif
			break;
	}
	return 0;
}

UpnpDevice_Handle handle = -1;

int main()
{
	int ret = UPNP_E_SUCCESS;
	int i;
	char desc_doc_url[200];

	i = read_config();
#ifdef DEBUG
	printf("Result: %d\n",i );
	if (ip_address)
		printf("Calling init: %s - %d\n", ip_address, port );
	else
		printf("Calling init: %d\n", port );
#endif
	ret = UpnpInit(ip_address, port);
	if (ret != UPNP_E_SUCCESS) {
		printf("Error in UpnpInit(): %d\n", ret);
		return ret;
	}

	ip_address = UpnpGetServerIpAddress();
	port = UpnpGetServerPort();
#ifdef DEBUG
	printf("http://%s:%d/%s\n", ip_address, port, "index.xml");
#endif
	snprintf(desc_doc_url, 200, "http://%s:%d/%s", ip_address, port, "index.xml");

	ret = UpnpSetWebServerRootDir("/etc/mserv");
	if (ret != UPNP_E_SUCCESS) {
		printf("Error in UpnpSetWebServerRootDir(): %d\n", ret);
		UpnpFinish();
		return ret;
	}

	ret = UpnpRegisterRootDevice(desc_doc_url,event,&handle,&handle);
	if (ret != UPNP_E_SUCCESS) {
		printf("Error in UpnpRegisterRootDevice(): %d\n", ret);
		UpnpFinish();
		return ret;
	}

	ret = UpnpSetVirtualDirCallbacks(&callbacks);
	ret = UpnpAddVirtualDir("/media");

	ret = UpnpSendAdvertisement(handle, 100);
	if (ret != UPNP_E_SUCCESS) {
		printf("Error in UpnpSendAdvertisement(): %d\n", ret);
		UpnpFinish();
		return ret;
	}

	while(1);
	return 0;
}
